from functools import partial
from concurrent.futures import ThreadPoolExecutor
from os import getcwd
import argparse

from rich.padding import PaddingDimensions
from bossman import Bossman
from bossman.repo import Revision
from bossman.abc import ResourceABC
from rich.panel import Panel
from rich.console import Console, RenderGroup
from rich.progress import Progress, TextColumn, BarColumn
from rich.table import Table
from rich.columns import Columns
from rich.prompt import Prompt

console = Console()

def init(subparsers: argparse._SubParsersAction):
  prerelease_parser = subparsers.add_parser("prerelease", help="prerelease applicable resources")
  prerelease_parser.add_argument("-e", "--exact-match", action="store_true", default=False, help="match resource exactly")
  prerelease_parser.add_argument("--yes", action="store_true", default=False, help="don't prompt")
  prerelease_parser.add_argument("--rev", required=False, default="HEAD", help="commit id or git ref to prerelease")
  prerelease_parser.add_argument("glob", nargs="*", default="*", help="select resources by glob pattern")
  prerelease_parser.set_defaults(func=exec, action="prerelease")

  release_parser = subparsers.add_parser("release", help="release applicable resources")
  release_parser.add_argument("-e", "--exact-match", action="store_true", default=False, help="match resource exactly")
  release_parser.add_argument("--yes", action="store_true", default=False, help="don't prompt")
  release_parser.add_argument("--rev", required=False, default="HEAD", help="commit id or git ref to release")
  release_parser.add_argument("glob", nargs="*", default="*", help="select resources by glob pattern")
  release_parser.set_defaults(func=exec, action="release")

def exec(bossman: Bossman, yes, rev, glob, exact_match:bool, action, *args, **kwargs):
  resources = bossman.get_resources(*glob, rev=rev, exact_match=exact_match)
  revision = bossman.get_revision(rev, resources)
  deployed, undeployed = [], []
  # a (pre)release operation should only be possible on resources that have been deployed by
  # a previous apply operation.
  for resource in resources:
    deployed.append(resource) if bossman.is_applied(resource, revision) else undeployed.append(resource)

  console.print("Preparing to {}:".format(action))
  console.print(Panel(RenderGroup(
    revision,
    "\n",
    Columns(deployed, expand=True, equal=True))
  ))
  if len(undeployed):
    console.print("\nThe following resources were selected but {} was not applied to them:\n".format(revision.id))
    console.print(Columns(undeployed, expand=True, equal=True))


  if not yes and not console.is_terminal:
    console.print("Input or output is not a terminal. Consider using --yes to forgo validation.")
    return
  if (not yes):
    if Prompt.ask("\nShall we proceed?", choices=("yes", "no"), default="no") != "yes":
      console.print("OK!")
      return

  with Progress(
      "[progress.description]{task.description}",
      BarColumn(bar_width=20),
      TextColumn("[progress.description]{task.fields[activation_status]}"),
    ) as progress_ui:

    from collections import defaultdict
    last_status = defaultdict(lambda: None)

    def on_update(task_id, resource: ResourceABC, status: str, progress: float):
      if isinstance(progress, (float, int)):
        progress_ui.start_task(task_id)
        progress = progress * 100
      progress_ui.update(task_id, completed=progress, activation_status=status)
      progress_ui.refresh()
      if not console.is_terminal:
        res_last_status = last_status.get(resource.path)
        if res_last_status != status:
          console.print(resource, status)
        last_status[resource.path] = status

    futures = []
    with ThreadPoolExecutor(100, "prereleasse") as executor:
      for resource in deployed:
        description = str(resource)
        if hasattr(resource, "__rich__"):
          description = resource.__rich__()
        task_id = progress_ui.add_task(description, total=100, activation_status="-", start=False)
        _on_update = partial(on_update, task_id)
        futures.append(executor.submit(getattr(bossman, action), resource, revision, _on_update))
      for resource, future in zip(deployed, futures):
        try:
          future.result()
        except Exception as e:
          print(":exclamation_mark:", resource, e)
          console.print_exception()




from .repo import Repo
import packaging.version
from rich.console import Console
from git import GitConfigParser

def __0_25_1(conf: GitConfigParser, repo: Repo, console: Console):
  """
  Until v0.25.1 inclusive, bossman init changed the git config to add explicit push/fetch
  refspecs to force git pull/push to also fetch/push notes.

  This behaviour was changed in favour of an explicit pull and push as part of application
  logic when relevant (e.g. before bossman status/prerelease/release, before an after bossman apply).

  The change was largely motivated by the fact that adding the fetch/push refspecs to the config
  appears to break normal branch tracking, forcing an explicit `git push origin ref`, not great UX.
  Because notes are critical to bossman, it's also better not to rely on the user remembering to push
  after apply.
  """
  notes_refspec = "+refs/notes/*:refs/notes/*"
  for section in conf.sections():
    if section.startswith("remote"):
      push_refspecs = conf.get_values(section, "push", [])
      if notes_refspec in push_refspecs:
        conf.remove_option(section, "push")
        push_refspecs = list(refspec for refspec in push_refspecs if refspec != notes_refspec)
        for refspec in push_refspecs:
          conf.add_value(section, "push", refspec)
        console.print(r"[red]-[/] \[{}] push: {}".format(section, notes_refspec))
      fetch_refspecs = conf.get_values(section, "fetch", [])
      if notes_refspec in fetch_refspecs:
        conf.remove_option(section, "fetch")
        fetch_refspecs = list(refspec for refspec in fetch_refspecs if refspec != notes_refspec)
        for refspec in fetch_refspecs:
          conf.add_value(section, "fetch", refspec)
        console.print(r"[red]-[/] \[{}] fetch: {}".format(section, notes_refspec))

MIGRATIONS = {
  "0.25.1": __0_25_1
}

def required(conf: GitConfigParser):
  parse = packaging.version.parse
  conf_version = parse(conf.get_value("bossman", "version", "0.0.0"))
  return any(conf_version <= parse(version) for version in MIGRATIONS)

def migrate(conf: GitConfigParser, repo: Repo, console: Console):
  if "bossman" not in conf.sections():
    conf.add_section("bossman")
  console.print("Initializing git configuration [yellow]{}[/]...".format(conf._file_or_files))
  conf_version = packaging.version.parse(conf.get_value("bossman", "version", "0.0.0"))
  for version, migration in MIGRATIONS.items():
    if conf_version <= packaging.version.parse(version):
      migration(conf, repo, console)


import sys
from bossman.errors import BossmanValidationError
from os import getcwd
import git
import argparse
from rich import print
from rich.table import Table
from bossman import Bossman

def init(subparsers: argparse._SubParsersAction):
  parser = subparsers.add_parser("validate", help="validates the working tree")
  parser.add_argument("-e", "--exact-match", action="store_true", default=False, help="match resource exactly")
  parser.add_argument("glob", nargs="*", default="*", help="select resources by glob pattern")
  parser.set_defaults(func=exec)

def exec(bossman: Bossman, glob, exact_match:bool, *args, **kwargs):
  resources = bossman.get_resources_from_working_copy(*glob, exact_match=exact_match)
  if len(resources):
    table = Table()
    table.add_column("Resource")
    table.add_column("Validation")
    table.add_column("Error")
    error_recorded = 0
    for resource in resources:
      error = ""
      try:
        bossman.validate(resource)
        status = ":thumbs_up:"
      except BossmanValidationError as e:
        status = ":thumbs_down:"
        error = e
        error_recorded = 1
      table.add_row(
        resource,
        status,
        error
      )
    print(table)
    if error_recorded == 1:
      sys.exit(1)
  else:
    print("No resources to show: check the glob pattern if provided, or the configuration.")




from collections import OrderedDict
from io import StringIO
import re
from bossman.repo import Revision


class GenericVersionComments:
  @staticmethod
  def from_revision(revision: Revision):
    comments = StringIO()
    comments.write(revision.message.strip())
    comments.write("\n\n")
    comments.write("commit: {}\n".format(revision.id))
    comments.write("branch: {}\n".format(", ".join(revision.branches)))
    comments.write("author: {} <{}>\n".format(revision.author_name, revision.author_email))
    if revision.author_email != revision.committer_email:
      comments.write("committer: {} <{}>\n".format(revision.commit.committer.name, revision.commit.committer.email))
    return GenericVersionComments(comments.getvalue())

  def __init__(self, comments: str):
    parts = comments.strip().rsplit("\n\n", 1)
    self.message = parts[0]
    self.metadata = (OrderedDict(re.findall(r"^([^:]+):\s*(.*)$", parts[1], re.MULTILINE))
      if len(parts) == 2
      else {})

  def __str__(self) -> str:
    return "{}\n\n{}".format(
      self.message,
      "\n".join(
        "{}: {}".format(k, v)
        for k, v in self.metadata.items()
      )
    )

  @property
  def commit(self):
    return self.metadata.get("commit", None)

  @property
  def subject_line(self):
    # 72 chars matches GitHub guidance, but it's a bit long for the console
    return self.message.split("\n")[0][:40]

  @property
  def branch(self):
    return self.metadata.get("branch", None)

  @property
  def author(self):
    return self.metadata.get("author", None)

  @property
  def committer(self):
    return self.metadata.get("committer", None)



import pathlib
from bossman.abc import ResourceABC

class SharedPolicyResource(ResourceABC):
  def __init__(self, path, **kwargs):
    super(SharedPolicyResource, self).__init__(path)
    self.__name = kwargs.get("name")

  @property
  def name(self):
    return self.__name

  @property
  def policy_path(self):
    # All operations use unix-style paths; this is important
    return str(pathlib.PurePosixPath(self.path) / "policy.json")

  @property
  def paths(self):
    return (self.policy_path,)

  def __rich__(self):
    prefix = self.path.replace(self.name, "")
    return "[grey53]{}[/][yellow]{}[/]".format(prefix, self.name)
